#toc h3 {
    margin-bottom: 0.1em;
}

#toc {
    margin-top: 3em;
    display: flex;
    flex-wrap: wrap;
}

#toc .item {
    padding-bottom: 1em;
    padding-right: 0.5em;

    flex: 1 1 auto;
    width: 230px;
}

#toc .description {
    margin: 0px;
    padding: 0px;
    margin-bottom: 0.5em;
}

#toc .links {
    margin: 0.1em;
    padding: 0.1em;
    font-size: 0.8em;
}

/* override table width restrictions */
.wy-table-responsive table td, .wy-table-responsive table th {
    /* !important prevents the common CSS stylesheets from
       overriding this as on RTD they are loaded after this stylesheet */
    white-space: normal !important;
}

.wy-table-responsive {
    overflow: visible !important;
}

/* small tweak for overview images */
#overview table img {
    margin-right: 30px;
    float: left;
}

#overview #key-features .wy-table-responsive:nth-of-type(1) table td {
    font-size: 1.2em;
}

/* tweak to limit width of column for settings default values */

.wy-table-responsive table tr td:nth-child(2), .wy-table-responsive tr table th:nth-child(2) {
    white-space: -pre-wrap !important;
}

#settings table tr td:nth-child(2) code {
    max-width: 100px;
    overflow-wrap: break-word;
    word-break: break-word;
    word-wrap: break-word;
    white-space: pre-wrap;
}